import math
from functools import wraps

import cv2
import numpy as np
import scipy.ndimage as ndi
from scipy.stats import median_abs_deviation
from skimage.transform import rescale, rotate


def get_contours(mask):
    mask = ((mask > 0) * 255).astype("uint8")
    contours, hierarchy = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    max_area = -1
    for i in range(len(contours)):
        area = cv2.contourArea(contours[i])
        if area > max_area:
            cnt = contours[i]
            max_area = area

    return cnt


def mask_coordinates(mask):
    cnt = get_contours(mask)

    x, y, w, h = cv2.boundingRect(cnt)
    return (x, y, w, h)


def crop_mask(mask):
    x, y, w, h = mask_coordinates(mask)
    return mask[y : y + h, x : x + w]


def get_area(img, name):
    mask = create_mask(img, name=name)
    cnt = get_contours(mask)
    return cv2.contourArea(cnt)


def get_scale(im1, im2):
    area1 = get_area(im1, name="1_" + im1.name)
    area2 = get_area(im2, name="2_" + im2.name)
    scale = math.sqrt(area1 / area2)
    return scale


def create_mask_v0(im):
    img = ndi.gaussian_filter(im, (3, 3))
    m = img.flatten()

    z1 = np.median(m[m > 0])
    z2 = median_abs_deviation(m[m > 0])
    mask = img

    mask = mask > z1 + 10 * z2
    img[~mask] = 0
    z = np.percentile(img.ravel(), 50)
    mask = mask > z

    mask = ndi.binary_opening(mask, iterations=3)

    mask = (
        ndi.binary_dilation(mask, iterations=3) * 1
        - ndi.binary_erosion(mask, iterations=3) * 1
    )
    mask = ndi.binary_fill_holes(mask)

    mask = mask * 255

    mask = ndi.gaussian_filter(mask, 3)
    mask = mask / mask.max() * 255

    mask = mask.astype("uint8")

    analysis = cv2.connectedComponentsWithStats(mask, 4, cv2.CV_16U)
    (totalLabels, label_ids, values, centroid) = analysis

    output = np.zeros(img.shape, dtype="uint8")
    area = 0
    for i in range(1, totalLabels):
        if values[i, cv2.CC_STAT_AREA] > area:
            area = values[i, cv2.CC_STAT_AREA]
            componentMask = (label_ids == i).astype("uint8") * 255

    # Creating the Final output mask
    output = cv2.bitwise_or(output, componentMask)

    output = ndi.binary_erosion(output, iterations=3)
    output = ((output > 0) * 255).astype("uint8")

    mask = im * output
    z1 = np.percentile(mask.ravel(), 50)
    mask = (mask > z1) * 255
    return mask.astype("uint8")


def maskcache(f):
    cache = {}

    @wraps(f)
    def wrapper(*args, **kwargs):
        name = kwargs.get("name", None)
        img = args[0]
        if name is None:
            return f(img.values, **kwargs)
        if name not in cache:
            cache[name] = f(img.values, **kwargs)
        # print(name)
        return cache[name]

    return wrapper


@maskcache
def create_mask(
    image,
    name=None,
    cl_morph_parm=10,
    c_min_area=0.02,
    c_count_max=15,
):

    """Create mask from the scaled image

    Parameters
    ----------
    image
        input image (better to be an scaled image)
    cl_morph_parm
        cv2.MORPH_ELLIPSE parameter for the Morphological closing
    c_min_area
        min area (relative to the full field area) for a contour to be considered as mask
    c_count_max
        Maximum number of contours to be checked for being mask
    output_size
        output image size

    Returns
    -------
    output
        Mask image with mask values set to 1 and background values set to 0
    """

    # to avoid compressing all background values around 0, and dealing with
    # zero values for background, we scale data to 25 and and add 1 later
    image = cv2.normalize(image, None, 1, 255, cv2.NORM_MINMAX, -1)

    # apply close morphology to get rid of small structures and make the big blog bolder
    outer = cv2.morphologyEx(
        image,
        cv2.MORPH_CLOSE,
        cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (cl_morph_parm, cl_morph_parm)),
    )

    # fill/replace zeros with median value to have smoother background
    # then we apply blur filter to remove any artificial casued by filling with median
    outer[outer <= 1] = np.median(outer[outer > 1])
    outer = outer.astype(np.uint8)
    outer = cv2.blur(outer, (5, 5))

    # remove noise by filtering the most frequent element from the image
    ret, th1 = cv2.threshold(
        outer, np.median(outer), 255, cv2.THRESH_BINARY + cv2.THRESH_TRIANGLE
    )

    # binaries the image and fill the holes
    image_fill_holes = ndi.binary_fill_holes(th1).astype(np.uint8)

    cnts, _ = cv2.findContours(
        image_fill_holes, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE
    )

    # create an empty mask image
    mask = np.zeros(image.shape, np.uint8)

    # assuming the min area for tissue is (e.g. 0.01 = 0.1x * 0.1y)
    min_blob_area = c_min_area * image.shape[1] * image.shape[0]
    blob_counter = 0

    # sort contours by area (large to small) and then loop over them and
    # select those with specific characteristics
    cnts = sorted(cnts, key=cv2.contourArea, reverse=True)
    for cnt in cnts[: np.min([int(c_count_max), len(cnts) - 1])]:
        if cv2.contourArea(cnt) > min_blob_area:
            blob_counter += 1
            cv2.drawContours(mask, [cnt], -1, 255, cv2.FILLED)
            mask = cv2.bitwise_and(image_fill_holes, mask)

    return (mask * 255).astype("uint8")


def get_flip(im_stpt, im_axio):
    mask_stpt = create_mask(im_stpt, name="1_" + im_stpt.name)
    mask_axio = create_mask(im_axio, name="2_" + im_axio.name)

    moments = cv2.moments(mask_stpt)
    hu_stpt = cv2.HuMoments(moments)

    moments = cv2.moments(mask_axio)
    hu_axio = cv2.HuMoments(moments)
    if np.sign(hu_stpt[-1]) == np.sign(hu_axio[-1]):
        return 0

    moments = cv2.moments(mask_axio[::-1, :])
    hu_axio = cv2.HuMoments(moments)
    if np.sign(hu_stpt[-1]) == np.sign(hu_axio[-1]):
        return 1

    moments = cv2.moments(mask_axio[:, ::-1])
    hu_axio = cv2.HuMoments(moments)
    if np.sign(hu_stpt[-1]) == np.sign(hu_axio[-1]):
        return 2


def func_iou(x, img1, img2):
    angle = x[0]
    scale = x[1]
    img2_rot = rotate(img2, angle, resize=True)
    img2_scl = rescale(img2_rot, scale)
    img2_shift = img2_scl
    # img2_shift = shift(img2_scl, (y_offset, x_offset))
    img2 = crop_mask(img2_shift)

    dy = max(img1.shape[0], img2.shape[0])
    dx = max(img1.shape[1], img2.shape[1])
    img1_f = np.pad(
        img1,
        [
            ((dy - img1.shape[0]) // 2, (dy - img1.shape[0]) // 2),
            ((dx - img1.shape[1]) // 2, (dx - img1.shape[1]) // 2),
        ],
    )

    img2_f = np.pad(
        img2,
        [
            ((dy - img2.shape[0]) // 2, (dy - img2.shape[0]) // 2),
            ((dx - img2.shape[1]) // 2, (dx - img2.shape[1]) // 2),
        ],
    )

    dy = min(img1_f.shape[0], img2_f.shape[0])
    dx = min(img1_f.shape[1], img2_f.shape[1])

    c1 = img1_f[:dy, :dx] > 0
    c2 = img2_f[:dy, :dx] > 0
    intersection = c1 * c2
    union = (c1 + c2) > 0
    iou = intersection.sum() / float(union.sum())
    return -np.log(iou)


def func_iou_old(x, *, img1, img2):
    angle = x[0]
    scale = x[1]
    # x_offset = x[2]
    # y_offset = x[3]
    img2_rot = rotate(img2, angle, resize=True)
    img2_scl = rescale(img2_rot, scale)
    # img2_shift = shift(img2_scl, (y_offset, x_offset))
    img2_f = crop_mask(img2_scl)
    dy = min(img1.shape[0], img2_f.shape[0])
    dx = min(img1.shape[1], img2_f.shape[1])
    c1 = img1[:dy, :dx] > 0
    c2 = img2_f[:dy, :dx] > 0
    intersection = c1 * c2
    union = (c1 + c2) > 0
    iou = intersection.sum() / float(union.sum())
    return 1 - iou



import os
import time
import traceback
from functools import partial, wraps
from multiprocessing import Pool

import cv2
import msgpack
import msgpack_numpy as mn
import numpy as np
import pandas as pd
import scipy.ndimage as ndi
import SimpleITK as sitk
import xarray as xr
from scipy import optimize
from scipy.stats import median_abs_deviation
from skimage.transform import rescale

from .contourlib import (
    create_mask,
    crop_mask,
    func_iou,
    get_contours,
    get_flip,
    get_scale,
)
from .plotting import NBPlot
from .registration import find_rigid_alignment, get_image_bb, get_initial_transform
from .terminal import debug, error, info, warn, set_verbose

os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"


def stats(im2, im1):
    """Compare two images using Hu moments and return statistic

    Parameters
    ----------
    im1, im2 : np.ndarray
       input images

    Returns
    -------
    comparison statistic (the smaller the more similar the images are)
    """

    mask1 = create_mask(im1, name="1_" + im1.name)
    cnt1 = get_contours(mask1)

    mask2 = create_mask(im2, name="2_" + im2.name)
    cnt2 = get_contours(mask2)

    ret = cv2.matchShapes(cnt1, cnt2, 1, 0.0)
    return ret


def match_image(ref_image, images):
    """Compute best matching image to a reference image

    Parameters
    ----------
    ref_image : np.ndarray
        Reference image
    images : list of np.ndarray or dask arrays
        List of images to compare to

    Returns
    -------
    index of most similar image in the input list
    """

    match = {}
    for j, img in enumerate(images):
        ret = stats(ref_image, img)
        match[j] = ret

    return match


def fit_angle_scale(im1, im2, scale=1, flip=0, use_gradients=False):
    mask1 = create_mask(im1, name="1_" + im1.name)
    mask2 = create_mask(im2, name="2_" + im2.name)

    if flip == 1:
        mask2 = mask2[::-1, :]
    elif flip == 2:
        mask2 = mask2[:, ::-1]

    mask2 = rescale(mask2, scale)

    mask1 = crop_mask(mask1)
    mask2 = crop_mask(mask2)

    if use_gradients:
        guess1 = angle_from_gradients(mask1, mask2)
        guess2 = guess1 + 180
        guess2 = guess2 - 360 * (guess2 > 360)
        out = []
        for angle in [guess1, guess2]:
            x = np.array([angle, 1])
            lsq = optimize.least_squares(
                func_iou,
                x,
                args=(mask2, mask1),
                diff_step=[1, 0.005],
                bounds=[(angle - 20, 0.96), (angle + 20, 1.04)],
            )
            out.append(lsq)
    else:
        out = []
        for angle in range(-140, 180, 40):
            x = np.array([angle, 1])
            lsq = optimize.least_squares(
                func_iou,
                x,
                args=(mask2, mask1),
                diff_step=[1, 0.005],
                bounds=[(angle - 40, 0.96), (angle + 40, 1.04)],
            )
            out.append(lsq)

    j = np.argmin([o.fun for o in out])
    angle = out[j].x[0]
    scale = scale * out[j].x[1]
    cost = out[j].cost
    return angle, scale, cost


def angle_from_gradients(mask1, mask2):
    def _corr(i, h, h2):
        return sum(h * np.roll(h2, i))

    # TODO: use ndi.distance_transform_edt ?
    t1 = ndi.distance_transform_bf(mask1)
    t2 = ndi.distance_transform_bf(mask2)

    t1 = t1.astype("float32")
    gx = cv2.Sobel(t1, cv2.CV_32F, 1, 0, ksize=-1)
    gy = cv2.Sobel(t1, cv2.CV_32F, 0, 1, ksize=-1)
    mag, angle1 = cv2.cartToPolar(gx, gy, angleInDegrees=True)
    angle1 = angle1.flatten()

    t2 = t2.astype("float32")
    gx = cv2.Sobel(t2, cv2.CV_32F, 1, 0, ksize=-1)
    gy = cv2.Sobel(t2, cv2.CV_32F, 0, 1, ksize=-1)
    mag, angle2 = cv2.cartToPolar(gx, gy, angleInDegrees=True)
    angle2 = angle2.flatten()

    h1, _ = np.histogram(angle1[angle1 > 0], bins=np.arange(0, 360))
    h2, _ = np.histogram(angle2[angle2 > 0], bins=np.arange(0, 360))

    # TODO: There must be a one liner for correlation of two vectors...
    c = np.array([_corr(i, h1, h2) for i in range(0, 359)])
    return np.argmax(c)


def safe(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        watch = time.monotonic()
        try:
            return f(*args, **kwargs)
        except Exception:
            error("ERROR:", traceback.format_exc())
        finally:
            debug("Elapsed", time.monotonic() - watch)

    return wrapper


"""

def func1(axio_entry_0, all_stpt_flip_0):
    -> list[(id, prob), (id, prob)]

def func2(axio_slice, stpt):
    for each id in axio_slice:
        func1(id, all_stpt)
    -> get the best

"""


@safe
def axio2stpt_firstpass(stpt_path, axio_path, section_axio, level="l.16", pl=None):

    ds_stpt = xr.open_zarr(f"{stpt_path}/mos", group=level)
    ds_axio = xr.open_zarr(f"{axio_path}/mos", group=level)

    # Prepare the data
    im_axio = ds_axio[section_axio].sel(z=0)
    im_mean_axio = im_axio.mean("channel")
    im_median_axio = im_axio.median("channel")
    if im_mean_axio.values.max() == 0:
        return

    images_stpt = [ds_stpt[s].sel(z=0).mean("channel") for s in list(ds_stpt)]

    # Get best match based on moments
    info("Get closest image based on moments", section_axio)
    match_st = match_image(im_median_axio, images_stpt)
    st_sorted = sorted(match_st.items(), key=lambda x: x[1])
    best_j = st_sorted[0][0]
    info("Best match", list(ds_stpt)[best_j])
    n_best = min(20, len(st_sorted) // 4)
    st_selected = [s[0] for s in st_sorted[:n_best]]

    """
        if method==1:
            # 1. load the fsb from stpt_path and axio_path
            # 2. for each axio and z

            # best_stpt gives
            top_n_stpt_list = func2(axio, stpt)
            # axio: S001 is stpt: S023

            for top_n_stpt_list:
                scale, etc. = func3_affine_transform()
                best_stpt
                fixed_axio = func4_transform(axio_im, stpt_im)

        elif method==2:
    """

    # Compute initial scale
    info("Compute initial scale", section_axio)
    section_stpt = list(ds_stpt)[best_j]
    im_stpt = ds_stpt[section_stpt].sel(z=0)
    im_mean_stpt = im_stpt.mean("channel")
    # im_mean_axio = im_mean_axio.compute()
    scale = get_scale(im_mean_stpt, im_mean_axio)

    # See if images are flipped
    info("Check if image is flipped", section_axio)
    flip = get_flip(im_mean_stpt, im_mean_axio)

    # Compute angle and improved scale
    info("Compute initial angle", section_axio)

    angle, scale, cost = fit_angle_scale(
        im_mean_stpt, im_mean_axio, scale=scale, flip=flip, use_gradients=False
    )

    debug(
        f"Initial transformation - angle: {angle:.2f}, scale: {scale:.2f}, flip: {flip:d}"
    )

    # Get initial transformation
    info("Compute initial transformation", section_axio)
    tx, fp1, fp2 = get_initial_transform(
        im_mean_stpt, im_mean_axio, angle=angle, scale=scale, flip=flip
    )

    info("Bead analysis", section_axio)
    cat_axio = pd.read_parquet(f"{axio_path}/bead/{section_axio}.parquet")
    x, y = cat_axio["x"] / 16, cat_axio["y"] / 16
    if flip == 1:
        y = len(ds_axio[section_axio].y) - y
    elif flip == 2:
        x = len(ds_axio[section_axio].x) - x
    x, y = x - fp2[0] + 100, y - fp2[1] + 100

    inv = tx._transform.GetInverse()
    c2 = np.array([inv.TransformPoint((xi, yi)) for xi, yi in zip(x, y)])

    nbeads = 0
    best_j2 = 0
    dbeads = {}
    for j in st_selected:
        if (j < 0) or (j > len(list(ds_stpt))):
            continue
        try:
            this = list(ds_stpt)[j]
        except Exception:
            continue

        cat_stpt = pd.read_parquet(f"{stpt_path}/bead/{this}.parquet")
        x, y = cat_stpt["x"] / 16, cat_stpt["y"] / 16
        x, y = x - fp1[0] + 100, y - fp1[1] + 100
        c1 = np.array([x, y]).T

        mc1 = c1.copy()

        if np.any(np.isnan(c1)):
            warn("Catalogue contains NaN values", this)

        for i in range(300):
            tc1 = []
            tc2 = []

            for c in mc1:
                res = np.sqrt((np.sum((c2 - c) ** 2, axis=1)))
                k = np.argmin(res)
                if res[k] < 200 / 16.0:
                    tc2.append(c2[k])
                    tc1.append(c)

            tc1 = np.array(tc1)
            tc2 = np.array(tc2)
            if len(tc1) < 10:
                break
            R, T = find_rigid_alignment(tc2, tc1)
            mc1 = np.array([R @ c + T for c in mc1])

        this_beads = len(tc1)
        if this_beads > nbeads:
            nbeads = this_beads
            best_j2 = j
        dbeads[this] = this_beads
        debug("Detected common beads", this, this_beads)

    info("Compute final transformation", section_axio)
    best_j = best_j2
    section_stpt = list(ds_stpt)[best_j]
    info("Best final match", section_stpt)
    im_stpt = ds_stpt[section_stpt].sel(z=0)
    im_mean_stpt = im_stpt.mean("channel")
    tx, fp1, fp2 = get_initial_transform(
        im_mean_stpt, im_mean_axio, angle=angle, scale=scale, flip=flip
    )

    if pl is not None:
        info("Plotting", section_axio)
        im1, im2, fp1, fp2 = get_image_bb(im_mean_stpt, im_mean_axio, flip=flip)

        indx = im2.ravel() > 0
        z1, z2 = np.median(im2.ravel()[indx]), median_abs_deviation(im2.ravel()[indx])
        im2 = im2.clip(0, z1 + 5 * z2)

        indx = im1.ravel() > 0
        z1, z2 = np.median(im1.ravel()[indx]), median_abs_deviation(im1.ravel()[indx])
        im1 = im1.clip(0, z1 + 5 * z2)

        # im1 = match_histograms(im1, im2)

        fixed = sitk.GetImageFromArray(im1)
        moving = sitk.GetImageFromArray(im2)
        outTx = tx._transform

        resampler = sitk.ResampleImageFilter()
        resampler.SetReferenceImage(fixed)
        resampler.SetInterpolator(sitk.sitkLinear)
        resampler.SetDefaultPixelValue(1)
        resampler.SetTransform(outTx)

        out = resampler.Execute(moving)
        array = sitk.GetArrayFromImage(out)

        data = (section_axio, section_stpt, match_st, dbeads, best_j, im1, array)
        data_enc = msgpack.packb(data, default=mn.encode)
        pl.send(data_enc)

    df = pd.DataFrame(
        [
            [section_axio, section_stpt, scale, flip, angle]
            + [fp2[0] - 100, fp2[1] - 100, fp2[2] + 200, fp2[3] + 200]
            + [fp1[0] - 100, fp1[1] - 100, fp1[2] + 200, fp1[3] + 200]
            + list(tx.center)
            + list(tx.translation)
            + list(tx.matrix)
        ],
        columns=[
            "axio",
            "stpt",
            "scale",
            "flip",
            "angle",
            "x_axio",
            "y_axio",
            "h_axio",
            "w_axio",
            "x_stpt",
            "y_stpt",
            "h_stpt",
            "w_stpt",
            "x_centre",
            "y_centre",
            "x_translation",
            "y_translation",
            "m00",
            "m10",
            "m01",
            "m11",
        ],
    )

    return df


def axio2stpt(args):

    if not args.stpt.exists():
        raise FileNotFoundError(f"File does not exist: {args.stpt}")

    if not args.axio.exists():
        raise FileNotFoundError(f"File does not exist: {args.axio}")

    ds_axio = xr.open_zarr(f"{args.axio}/mos")

    pl = NBPlot(args.pdf)

    info(f"Using {args.nprocs} processes")
    if args.sections is not None:
        sections = args.sections.split(",")
    else:
        sections = list(ds_axio)

    info(f"Running for {len(sections)} sections")
    watch = time.monotonic()

    set_verbose(args.verbose)

    f = partial(axio2stpt_firstpass, args.stpt, args.axio, pl=pl.plot_pipe)
    with Pool(processes=args.nprocs) as pool:
        df = pool.map(f, sections)

    pl.plot(finished=True)
    pl.plot_process.join()
    try:
        df = pd.concat(df)
    except Exception as e:
        error("ERROR:", e)
        df = pd.DataFrame()

    if args.output.suffix.endswith("csv"):
        df.to_csv(f"{args.output}", index=False)
    elif args.output.suffix.endswith("parquet"):
        df.to_parquet(f"{args.output}", index=False)
    elif args.output.suffix.endswith("npy"):
        df.to_numpy(f"{args.output}", index=False)
    elif args.output.suffix.endswith("pkl"):
        df.to_pkl(f"{args.output}", index=False)
    elif args.output == "stdout":
        print(df.to_string())
    else:
        df.to_csv(f"{args.output}", index=False)

    elapsed = int(time.monotonic() - watch)
    info(f"Finished in {elapsed} seconds")



from colored import fg, bg, attr
import multiprocessing as mp

VERBOSE = True


def set_verbose(val):
    global VERBOSE
    VERBOSE = val


def debug(*args):
    if not VERBOSE:
        return
    curr_proc = mp.current_process()
    print(f"{curr_proc.name}:{fg('blue')}", *args, attr("reset"))


def info(*args):
    if not VERBOSE:
        return
    curr_proc = mp.current_process()
    print(f"{curr_proc.name}:", *args)


def warn(*args):
    if not VERBOSE:
        return
    curr_proc = mp.current_process()
    print(f"{curr_proc.name}:{fg('magenta')}", *args, attr("reset"))


def error(*args):
    if not VERBOSE:
        return
    curr_proc = mp.current_process()
    print(f"{curr_proc.name}:{fg('white')}{bg('red')}", *args, attr("reset"))



import multiprocessing as mp

import msgpack
import msgpack_numpy as mn
import numpy as np


class ProcessPlotter:
    def __init__(self, output):
        self.output = output

    def terminate(self):
        pass

    def call_back(self):
        import matplotlib.gridspec as gridspec
        import matplotlib.pyplot as plt
        from matplotlib.backends.backend_pdf import PdfPages

        self.flag = False

        with PdfPages(f"{self.output}") as pdf:
            while True:
                command = self.pipe.recv()
                if command is None:
                    self.terminate()
                    return False
                else:
                    (
                        section_axio,
                        section_stpt,
                        match_st,
                        dbeads,
                        best_j,
                        im1,
                        array,
                    ) = msgpack.unpackb(
                        command,
                        object_hook=mn.decode,
                        use_list=False,
                        strict_map_key=False,
                    )
                    fig = plt.figure(tight_layout=True, figsize=(8, 12))
                    gs = gridspec.GridSpec(4, 2)

                    ax = fig.add_subplot(gs[0, :])
                    ax.plot(match_st.values())
                    ax.plot([best_j], [match_st[best_j]], "o")
                    ax.set_xlabel("STPT Slice No.")
                    ax.set_ylabel("stat")
                    ax.set_title(f"A: {section_axio} S: {section_stpt}")

                    ax = fig.add_subplot(gs[1, :])
                    vals = sorted(list(dbeads.items()), key=lambda x: x[0])
                    slices = [x[0] for x in vals]
                    nbeads = [x[1] for x in vals]
                    j = np.argmax(nbeads)
                    j = max(j, 5)
                    ax.bar(slices[j - 5 : j + 5], nbeads[j - 5 : j + 5])
                    ax.set_ylabel("beads")

                    ax = fig.add_subplot(gs[2:, :])
                    img = np.zeros((im1.shape[0], im1.shape[1], 3))
                    img[..., 0] = (im1 - im1.min()) / im1.max()
                    img[..., 1] = (array - array.min()) / array.max()

                    ax.imshow(img)
                    ax.axis("off")

                    pdf.savefig()
                    plt.close()
        return True

    def __call__(self, pipe):
        self.pipe = pipe
        self.call_back()


class NBPlot:
    def __init__(self, output):
        self.plot_pipe, plotter_pipe = mp.Pipe()
        self.plotter = ProcessPlotter(output)
        self.plot_process = mp.Process(
            target=self.plotter, args=(plotter_pipe,), daemon=True
        )
        self.plot_process.start()

    def plot(self, *args, finished=False):
        send = self.plot_pipe.send
        if finished:
            send(None)
        else:
            send(args)



import sys
from argparse import ArgumentParser, Namespace
from typing import List
from pathlib import Path


from .axio2stpt import axio2stpt

__version__ = "0.2.0"
__author__ = "E.A. Gonzales Solares"


def parse_args(input: List[str]) -> Namespace:

    parent_parser = ArgumentParser(add_help=False)

    main_parser = ArgumentParser(description="IMAXT Registration")
    main_parser.add_argument(
        "-V", "--version", action="version", version="%(prog)s (" + __version__ + ")"
    )

    subparsers = main_parser.add_subparsers(help="commands", dest="command")

    # axio2stpt
    a2s_parser = subparsers.add_parser("axio2stpt", parents=[parent_parser])
    a2s_parser.add_argument(
        "--stpt", required=True, type=Path, help="Full path to STPT dataset"
    )
    a2s_parser.add_argument(
        "--axio", required=True, type=Path, help="Full path to AXIO dataset"
    )
    a2s_parser.add_argument(
        "--output",
        required=True,
        type=Path,
        help="Destination of output correspondence table (csv, parquet, npy, pkl)",
    )
    a2s_parser.add_argument(
        "--pdf",
        required=True,
        type=Path,
        help="Destination of outout PDF quality control file",
    )
    a2s_parser.add_argument(
        "--verbose", action="store_true", default=False, help="Verbose output"
    )
    a2s_parser.add_argument(
        "--nprocs",
        required=False,
        type=int,
        default=6,
        help="Number of processes to run in parallel (default 6)",
    )
    a2s_parser.add_argument(
        "--sections",
        required=False,
        type=str,
        default=None,
        help="Comma separated list of AXIO sections to do (default is all)",
    )
    a2s_parser.set_defaults(func=axio2stpt)

    args = main_parser.parse_args(input)
    return args


def main():
    """Main entry point for owl.

    Invoke the command line help with::

    """
    args = parse_args(sys.argv[1:])

    if hasattr(args, "func"):
        args.func(args)



import SimpleITK as sitk
from skimage.transform import rotate, rescale
import math
import numpy as np
from typing import Tuple
from scipy.stats import median_abs_deviation
from skimage.exposure import match_histograms

from .contourlib import mask_coordinates, create_mask


class Transformation(dict):

    __slots__ = ["metric", "translation", "matrix", "center", "transform"]

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):  # pragma: nocover
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return "\n".join(
                [k.rjust(m) + ": " + repr(v) for k, v in sorted(self.items())]
            )
        else:
            return self.__class__.__name__ + "()"

    def __dir__(self):
        return list(self.keys())


def _command_iteration(method):
    if method.GetOptimizerIteration() == 0:
        print("Estimated Scales: ", method.GetOptimizerScales())
    print(
        f"{method.GetOptimizerIteration():3} "
        + f"= {method.GetMetricValue():7.5f} "
        + f": {method.GetOptimizerPosition()}"
    )


def get_transformation(img1, img2, angle=0, scale=1, verbose=False):

    theta = angle / 180 * np.pi
    matrix = (
        np.array([math.cos(theta), math.sin(theta), -math.sin(theta), math.cos(theta)])
        / scale
    )

    at = sitk.AffineTransform(
        matrix, (0, 0), (img2.shape[0] // 2 + 0.5, img2.shape[1] / 2 + 0.5)
    )

    fixed = sitk.GetImageFromArray(img1)
    moving = sitk.GetImageFromArray(img2)

    R = sitk.ImageRegistrationMethod()
    R.SetMetricAsCorrelation()
    R.SetOptimizerAsRegularStepGradientDescent(
        learningRate=2.0,
        minStep=1e-4,
        numberOfIterations=1000,
        gradientMagnitudeTolerance=1e-8,
    )
    R.SetOptimizerScalesFromIndexShift()

    tx = sitk.CenteredTransformInitializer(fixed, moving, at)
    R.SetInitialTransform(tx)
    R.SetInterpolator(sitk.sitkLinear)
    if verbose:
        R.AddCommand(sitk.sitkIterationEvent, lambda: _command_iteration(R))

    outTx = R.Execute(fixed, moving)

    data = {
        "metric": R.GetMetricValue(),
        "translation": outTx.GetTranslation(),
        "matrix": outTx.GetMatrix(),
        "center": outTx.GetCenter(),
        "_transform": outTx,
    }

    return Transformation(data)


def get_image_bb(img1, img2, scale=1, angle=0, flip=0):

    mask1 = create_mask(img1, name="1_" + img1.name)
    im1 = mask1 * img1.values / mask1.max()

    mask2 = create_mask(img2, name="2_" + img2.name)
    im2 = mask2 * img2.values / mask2.max()

    im2 = rescale(im2, scale)
    mask2 = rescale(mask2, scale)
    if flip == 1:
        im2 = im2[::-1, :]
        mask2 = mask2[::-1, :]
    elif flip == 2:
        im2 = im2[:, ::-1]
        mask2 = mask2[:, ::-1]

    im2 = rotate(im2, angle, resize=False)
    mask2 = rotate(mask2, angle, resize=False)

    x1, y1, w1, h1 = mask_coordinates(mask1)
    x2, y2, w2, h2 = mask_coordinates(mask2)

    im1 = im1[y1 : y1 + h1, x1 : x1 + w1]
    im2 = im2[y2 : y2 + h2, x2 : x2 + w2]

    dy = dx = 200
    im1 = np.pad(im1, (dy // 2, dx // 2))
    im2 = np.pad(im2, (dy // 2, dx // 2))

    return im1, im2, (x1, y1, h1, w1), (x2, y2, h2, w2)


def get_initial_transform(im_mean1, im_mean2, flip=0, angle=0, scale=1):
    im1, im2, fp1, fp2 = get_image_bb(im_mean1, im_mean2, flip=flip)

    indx = im2.ravel() > 0
    z1, z2 = np.median(im2.ravel()[indx]), median_abs_deviation(im2.ravel()[indx])
    im2 = im2.clip(0, z1 + 10 * z2)

    indx = im1.ravel() > 0
    z1, z2 = np.median(im1.ravel()[indx]), median_abs_deviation(im1.ravel()[indx])
    im1 = im1.clip(0, z1 + 10 * z2)

    im1 = match_histograms(im1, im2)

    t = get_transformation(im1, im2, angle=angle, scale=scale, verbose=False)
    return t, fp1, fp2


def find_rigid_alignment(A: np.ndarray, B: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    2-D or 3-D registration with known correspondences.
    Registration occurs in the zero centered coordinate system, and then
    must be transported back.

    Args:
        A: Array of shape (N,D) -- Reference Point Cloud (target)
        B: Array of shape (N,D) -- Point Cloud to Align (source)

    Returns:
        R: optimal rotation (3,3)
        t: optimal translation (3,)
    """
    num_pts = A.shape[0]
    dim = A.shape[1]

    a_mean = np.median(A, axis=0)
    b_mean = np.median(B, axis=0)

    # Zero-center the point clouds
    A -= a_mean
    B -= b_mean

    N = np.zeros((dim, dim))
    for i in range(num_pts):
        N += A[i].reshape(dim, 1) @ B[i].reshape(1, dim)
    N = A.T @ B

    U, D, V_T = np.linalg.svd(N)
    S = np.eye(dim)
    det = np.linalg.det(U) * np.linalg.det(V_T.T)

    # Check for reflection case
    if not np.isclose(det, 1.0):
        S[dim - 1, dim - 1] = -1

    R = U @ S @ V_T
    t = R @ b_mean.reshape(dim, 1) - a_mean.reshape(dim, 1)
    return R, -t.squeeze()
